import os
import pickle
import numpy as np
import pandas as pd
from collections import defaultdict

def padding_users(vali_data, NUM_USERS, NUM_ITEMS):
    user_list = vali_data[:, 0]
    item_list = vali_data[:, 1]
    rating_list = vali_data[:, 2]
    
    tmp_u_item_dict = defaultdict(list)
    tmp_u_rating_dict = defaultdict(list)
    
    u_item_dict = defaultdict(list)
    u_item_length_dict = dict()
    
    for idx, u in enumerate(user_list):
        tmp_u_item_dict[u].append(item_list[idx])
        tmp_u_rating_dict[u].append(rating_list[idx])
    
    new_vali_data = []
    
    #import pdb; pdb.set_trace()
    max_length = 0
    for u in tmp_u_item_dict.keys():
        real_item_rating_list = tmp_u_rating_dict[u]
        sort_index = np.argsort(real_item_rating_list)
        sort_index = sort_index[::-1]
        
        for nxy_idx in sort_index:
            u_item_dict[u].append(tmp_u_item_dict[u][nxy_idx])
        
        u_item_length_dict[u] = len(u_item_dict[u])
        
        max_length = max(max_length, u_item_length_dict[u])
        
        final_idx = len(sort_index)
        positive_set = set(u_item_dict[u])
        while final_idx < 100:
            sample_item = np.random.randint(NUM_ITEMS)
            while sample_item in positive_set:
                sample_item = np.random.randint(NUM_ITEMS)
            
            u_item_dict[u].append(sample_item)
            
            positive_set.add(sample_item)
            final_idx = final_idx + 1
            
    return u_item_dict, u_item_length_dict, max_length

# choose dataset to process
dataset = 'steam-v1'
raw_data_path = os.path.join('/root/autodl-fs/pmf_torch/data/steam-v1/ratings.dat')


root_dir = '/root/autodl-fs/pmf_torch'
processed_data_path = os.path.join('/root/autodl-fs/pmf_torch/small/processed_data/', dataset)

names = ['user_id', 'item_id', 'rating', 'timestamp']
read = pd.read_csv(raw_data_path, sep='\t', nrows = 100000, names=names, header=1)

users = list(set(read['user_id']))
user_id_index = dict((user_id, index) for user_id, index in zip(users, range(len(users))))
items = list(set(read['item_id']))
item_id_index = dict((item_id, index) for item_id, index in zip(items, range(len(items))))

# iterator all users and find each users' max label and min label
user_rating_dict = defaultdict(list)
item_rating_dict = defaultdict(list)
with open(root_dir+'/data/'+dataset+'/ratings.dat', 'r') as f:
    lines = f.readlines()[:100000]
    #import pdb; pdb.set_trace()
    for i in range(1, len(lines)):
        line = lines[i].strip().split('\t')
        user_id = int(line[0])
        item_id = int(line[1])
        rating = float(line[2])
        user_rating_dict[user_id].append(rating)
        item_rating_dict[item_id].append(rating)

avg_user_rating_dict = dict()
avg_item_rating_dict = dict()
std_user_rating_dict = dict()
std_item_rating_dict = dict()

for user in user_rating_dict:
    avg_user_rating_dict[user] = np.mean(user_rating_dict[user])
    std_user_rating_dict[user] = np.std(user_rating_dict[user])

for item in item_rating_dict:
    avg_item_rating_dict[item] = np.mean(item_rating_dict[item])
    std_item_rating_dict[item] = np.std(item_rating_dict[item])


min_rating = 0
max_rating = 0
data = []
with open(root_dir+'/data/'+dataset+'/ratings.dat', 'r') as f:
    lines = f.readlines()[:100000]
    count_user = 0
    count_item = 0
    for i in range(1, len(lines)):
        line = lines[i].strip().split('\t')
        user_id = int(line[0])
        item_id = int(line[1])
        rating = float(line[2]) 
        #if std_user_rating_dict[user_id]:
            #import pdb; pdb.set_trace()
        rating_user = (rating - avg_user_rating_dict[user_id]) / (std_user_rating_dict[user_id] + 1e-6)
        rating_item = (rating - avg_item_rating_dict[item_id]) / (std_item_rating_dict[item_id] + 1e-6)
        rating = (rating_user + rating_item) / 2.0 / 6.0
        
        min_rating = min(min_rating, rating)
        max_rating = max(max_rating, rating)

        data.append([user_id_index[user_id], item_id_index[item_id], rating])

#import pdb; pdb.set_trace()

print(max_rating)
print(min_rating)

data = np.array(data)
np.random.shuffle(data)

pickle.dump(user_id_index, open(os.path.join(processed_data_path, 'user_id_index.pkl'), 'wb'))
pickle.dump(item_id_index, open(os.path.join(processed_data_path, 'item_id_index.pkl'), 'wb'))
np.savetxt(os.path.join(processed_data_path, 'data.txt'), data, fmt='%f')

NUM_USERS = max(user_id_index.values()) + 1
NUM_ITEMS = max(item_id_index.values()) + 1

# set split ratio
ratio = 0.8

train_data = data[:int(ratio*data.shape[0])]
vali_data = data[int(ratio*data.shape[0]):int((ratio+(1-ratio)/2)*data.shape[0])]
test_data = data[int((ratio+(1-ratio)/2)*data.shape[0]):]
#import pdb; pdb.set_trace()

f = open('/root/autodl-fs/pmf_torch/small/tmp_data/train_data.pkl', 'wb')
pickle.dump(train_data, f) 
f.close()

f = open('/root/autodl-fs/pmf_torch/small/tmp_data/vali_data.pkl', 'wb')
pickle.dump(vali_data, f) 
f.close()

f = open('/root/autodl-fs/pmf_torch/small/tmp_data/test_data.pkl', 'wb')
pickle.dump(test_data, f) 
f.close()

vali_u_item_dict, vali_u_item_length_dict, vali_max_length = padding_users(vali_data, NUM_USERS, NUM_ITEMS)
test_u_item_dict, test_u_item_length_dict, test_max_length = padding_users(test_data, NUM_USERS, NUM_ITEMS)

pickle.dump(vali_u_item_dict, open('/root/autodl-fs/pmf_torch/small/processed_data/steam-v1/valid_u_item_dict.pkl', 'wb'))
pickle.dump(vali_u_item_length_dict, open('/root/autodl-fs/pmf_torch/small/processed_data/steam-v1/valid_u_item_length_dict.pkl', 'wb'))

pickle.dump(test_u_item_dict, open('/root/autodl-fs/pmf_torch/small/processed_data/steam-v1/test_u_item_dict.pkl', 'wb'))
pickle.dump(test_u_item_length_dict, open('/root/autodl-fs/pmf_torch/small/processed_data/steam-v1/test_u_item_length_dict.pkl', 'wb'))